# FUNCTIONS THAT MODIFY RQ TO NOT COMPLAIN ABOUT SINGULAR DESIGNS AND TO WORK FOR SPARSE DESIGNS

rq.nostop <- function (formula, tau = 0.5, data, subset, weights, na.action, 
          method = "br", model = TRUE, contrasts = NULL, ...) 
{
  call <- match.call()
  mf <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data", "subset", "weights", "na.action"), 
             names(mf), 0)
  mf <- mf[c(1, m)]
  mf$drop.unused.levels <- TRUE
  mf[[1]] <- as.name("model.frame")
  mf <- eval.parent(mf)
  if (method == "model.frame") 
    return(mf)
  mt <- attr(mf, "terms")
  weights <- as.vector(model.weights(mf))
  Y <- model.response(mf)
  X <- model.matrix(mt, mf, contrasts)
  eps <- .Machine$double.eps^(2/3)
  Rho <- function(u, tau) u * (tau - (u < 0))
  if (length(tau) > 1) {
    if (any(tau < 0) || any(tau > 1)) 
      stop("invalid tau:  taus should be >= 0 and <= 1")
    if (any(tau == 0)) 
      tau[tau == 0] <- eps
    if (any(tau == 1)) 
      tau[tau == 1] <- 1 - eps
    coef <- matrix(0, ncol(X), length(tau))
    rho <- rep(0, length(tau))
    fitted <- resid <- matrix(0, nrow(X), length(tau))
    for (i in 1:length(tau)) {
      z <- {
        if (length(weights)) 
          rq.wfit.nostop(X, Y, tau = tau[i], weights, method, 
                  ...)                                         #change with respect to rq
        else rq.fit.nostop(X, Y, tau = tau[i], method, ...)    #change with respect to rq
      }
      coef[, i] <- z$coefficients
      if (length(z$residuals)) { #changed with respec to rq to allow for NA's
        resid[, i] <- z$residuals
        rho[i] <- sum(Rho(z$residuals, tau[i]))
        fitted[, i] <- Y - z$residuals
        }
    }
    taulabs <- paste("tau=", format(round(tau, 3)))
    dimnames(coef) <- list(dimnames(X)[[2]], taulabs)
    dimnames(resid) <- list(dimnames(X)[[1]], taulabs)
    fit <- z
    fit$coefficients <- coef
    fit$residuals <- resid
    fit$fitted.values <- fitted
    if (method == "lasso") 
      class(fit) <- c("lassorqs", "rqs")
    else if (method == "scad") 
      class(fit) <- c("scadrqs", "rqs")
    else class(fit) <- "rqs"
  }
  else {
    process <- (tau < 0 || tau > 1)
    if (tau == 0) 
      tau <- eps
    if (tau == 1) 
      tau <- 1 - eps
    fit <- {
      if (length(weights)) 
        rq.wfit.nostop(X, Y, tau = tau, weights, method, ...) #change
      else rq.fit.nostop(X, Y, tau = tau, method, ...)        #change
    }
    if (process) 
      rho <- list(x = fit$sol[1, ], y = fit$sol[3, ])
    else {
      dimnames(fit$residuals) <- list(dimnames(X)[[1]], 
                                      NULL)
      rho <- sum(Rho(fit$residuals, tau))
    }
    if (method == "lasso") 
      class(fit) <- c("lassorq", "rq")
    else if (method == "scad") 
      class(fit) <- c("scadrq", "rq")
    else class(fit) <- ifelse(process, "rq.process", "rq")
  }
  fit$na.action <- attr(mf, "na.action")
  fit$formula <- formula
  fit$terms <- mt
  fit$xlevels <- .getXlevels(mt, mf)
  fit$call <- call
  fit$tau <- tau
  fit$weights <- weights
  fit$residuals <- drop(fit$residuals)
  fit$rho <- rho
  fit$method <- method
  fit$fitted.values <- drop(fit$fitted.values)
  attr(fit, "na.message") <- attr(m, "na.message")
  if (model) 
    fit$model <- mf
  fit
}


rq.fit.nostop <- function (x, y, tau = 0.5, method = "br", ...) 
{
  if (length(tau) > 1) {
    tau <- tau[1]
    warning("Multiple taus not allowed in rq.fit: solution restricted to first element")
  }
  # change with respect to rq.fit
  fit <- switch(method, fn = rq.fit.fnb.nostop(x, y, tau = tau, ...), 
                fnb = rq.fit.fnb.nostop(x, y, tau = tau, ...), fnc = rq.fit.fnc(x, 
                                                                         y, tau = tau, ...), pfn = rq.fit.pfn.nostop(x, y, tau = tau, 
                                                                                                              ...), br = rq.fit.br(x, y, tau = tau, ...), sfn2 = rq.fit.sfn.nostop(x, y, tau = tau, ...), lasso = rq.fit.lasso(x, 
                                                                                                                                                                               y, tau = tau, ...), scad = rq.fit.scad(x, y, tau = tau, 
                                                                                                                                                                                                                      ...), {
                                                                                                                                                                                                                        what <- paste("rq.fit.", method, sep = "")
                                                                                                                                                                                                                        if (exists(what, mode = "function")) (get(what, mode = "function"))(x, 
                                                                                                                                                                                                                                                                                            y, ...) else stop(paste("unimplemented method:", 
                                                                                                                                                                                                                                                                                                                    method))
                                                                                                                                                                                                                      })
  fit$fitted.values <- y - fit$residuals
  fit$contrasts <- attr(x, "contrasts")
  fit
}

# Add the method "sfn2" to run quantile regressions with sparse designs

rq.wfit.nostop <- function (x, y, tau = 0.5, weights, method = "br", ...) 
{
  if (any(weights < 0)) 
    stop("negative weights not allowed")
  if (length(tau) > 1) {
    tau <- tau[1]
    warning("Multiple taus not allowed in rq.wfit: solution restricted to first element")
  }
  contr <- attr(x, "contrasts")
  wx <- x * weights
  wy <- y * weights
  # change with respect to rq.fit adding sfn2 method
  fit <- switch(method, fn = rq.fit.fnb.nostop(wx, wy, tau = tau, 
                                        ...), fnb = rq.fit.fnb.nostop(wx, wy, tau = tau, ...), br = rq.fit.br(wx, 
                                                                                                       wy, tau = tau, ...), fnc = rq.fit.fnc(wx, wy, tau = tau, 
                                                                                                                                             ...), pfn = rq.fit.pfn.nostop(wx, wy, tau = tau, ...), sfn2 = rq.fit.sfn.nostop(wx, wy, tau = tau, ...), {
                                                                                                                                               what <- paste("rq.fit.", method, sep = "")
                                                                                                                                               if (exists(what, mode = "function")) (get(what, mode = "function"))(x, 
                                                                                                                                                                                                                   y, ...) else stop(paste("unimplemented method:", 
                                                                                                                                                                                                                                           method))
                                                                                                                                             })
  if (length(fit$sol)) {                                         #change because original if/else does not work for sfn2
    fit$fitted.values <- x %*% fit$sol[-(1:3), ] 
  } else  { fit$fitted.values <- x %*% fit$coefficients
#            names(fit$coefficients) <- dimnames(x)[[2]]
    }
  fit$residuals <- y - fit$fitted.values
  fit$contrasts <- attr(x, "contrasts")
  fit$weights <- weights
  fit
}

# function that does not complain about singular designs and records instead a NaN



rq.fit.pfn.nostop <- function (x, y, tau = 0.5, Mm.factor = 0.8, max.bad.fixup = 3, 
          eps = 1e-06) 
{
  n <- length(y)
  if (nrow(x) != n) 
    stop("x and y don't match n")
  if (tau < 0 | tau > 1) 
    stop("tau outside (0,1)")
  p <- ncol(x)
  m <- round(((p + 1) * n)^(2/3))
  not.optimal <- TRUE
  while (not.optimal) {
    if (m < n) 
      s <- sample(n, m)
    else {
      b <- rq.fit.fnb.nostop(x, y, tau = tau, eps = eps)$coef   #change
      break
    }
    xx <- x[s, ]
    yy <- y[s]
    z <- rq.fit.fnb.nostop(xx, yy, tau = tau, eps = eps)  #change
    xxinv <- solve(chol(crossprod(xx)))
    band <- sqrt(((x %*% xxinv)^2) %*% rep(1, p))
    r <- y - x %*% z$coef
    M <- Mm.factor * m
    lo.q <- max(1/n, tau - M/(2 * n))
    hi.q <- min(tau + M/(2 * n), (n - 1)/n)
    kappa <- quantile(r/pmax(eps, band), c(lo.q, hi.q))
    sl <- r < band * kappa[1]
    su <- r > band * kappa[2]
    bad.fixup <- 0
    while (not.optimal & (bad.fixup < max.bad.fixup)) {
      xx <- x[!su & !sl, ]
      yy <- y[!su & !sl]
      if (any(sl)) {
        glob.x <- c(t(x[sl, , drop = FALSE]) %*% rep(1, 
                                                     sum(sl)))
        glob.y <- sum(y[sl])
        xx <- rbind(xx, glob.x)
        yy <- c(yy, glob.y)
      }
      if (any(su)) {
        ghib.x <- c(t(x[su, , drop = FALSE]) %*% rep(1, 
                                                     sum(su)))
        ghib.y <- sum(y[su])
        xx <- rbind(xx, ghib.x)
        yy <- c(yy, ghib.y)
      }
      z <- rq.fit.fnb.nostop(xx, yy, tau = tau, eps = eps)
      b <- z$coef
      r <- y - x %*% b
      su.bad <- (r < 0) & su
      sl.bad <- (r > 0) & sl
      if (any(c(su.bad, sl.bad))) {
        if (sum(su.bad | sl.bad) > 0.1 * M) {
          warning("Too many fixups:  doubling m")
          m <- 2 * m
          break
        }
        su <- su & !su.bad
        sl <- sl & !sl.bad
        bad.fixup <- bad.fixup + 1
      }
      else not.optimal <- FALSE
    }
  }
  coefficients <- b
  names(coefficients) <- dimnames(x)[[2]]
  residuals <- y - x %*% b
  return(list(coefficients = coefficients, tau = tau, residuals = residuals))
}

rq.fit.fnb.nostop <- function (x, y, tau = 0.5, beta = 0.99995, eps = 1e-06) 
{
  n <- length(y)
  p <- ncol(x)
  if (n != nrow(x)) 
    stop("x and y don't match n")
  if (tau < eps || tau > 1 - eps) 
    stop("No parametric Frisch-Newton method.  Set tau in (0,1)")
  rhs <- (1 - tau) * apply(x, 2, sum)
  d <- rep(1, n)
  u <- rep(1, n)
  wn <- rep(0, 10 * n)
  wn[1:n] <- (1 - tau)
  z <- .Fortran("rqfnb", as.integer(n), as.integer(p), a = as.double(t(as.matrix(x))), 
                c = as.double(-y), rhs = as.double(rhs), d = as.double(d), 
                as.double(u), beta = as.double(beta), eps = as.double(eps), 
                wn = as.double(wn), wp = double((p + 3) * p), it.count = integer(3), 
                info = integer(1), PACKAGE = "quantreg")
  if (z$info != 0) {
    coefficients <- rep(NA,p)
  }
  # stop(paste("Error info = ", z$info, "in stepy: singular design"))
  else {
    coefficients <- -z$wp[1:p]
  }
#  if (z$info != 0) 
#    stop(paste("Error info = ", z$info, "in stepy: singular design"))
#  coefficients <- -z$wp[1:p]
  names(coefficients) <- dimnames(x)[[2]]
  residuals <- y - x %*% coefficients
  list(coefficients = coefficients, tau = tau, residuals = residuals)
}

# modified rq.fit.sfn that transform design matrix into csr matrix 

rq.fit.sfn.nostop <- function (a, y, tau = 0.5, rhs = (1 - tau) * c(t(a) %*% rep(1, 
                                                            length(y))), control) 
{
  a <- as.matrix.csr(a)    # change with respect to rq.fit.sfn
  y <- -y
  n <- length(y)
  m <- a@dimension[2]
  if (n != a@dimension[1]) 
    stop("Dimensions of design matrix and the response vector not compatible")
  u <- rep(1, length = n)
  x <- rep((1 - tau), length = n)
  nnzdmax <- nnza <- a@ia[n + 1] - 1
  iwmax <- 7 * m + 3
  ao <- t(a)
  e <- ao %*% a
  nnzemax <- e@ia[m + 1] - 1
  ctrl <- sfn.control()
  if (!missing(control)) {
    control <- as.list(control)
    ctrl[names(control)] <- control
  }
  nsubmax <- ctrl$nsubmax
  tmpmax <- ctrl$tmpmax
  nnzlmax <- ctrl$nnzlmax
  if (is.null(ctrl$nsubmax)) 
    nsubmax <- nnzemax
  if (is.null(ctrl$tmpmax)) 
    tmpmax <- 6 * m
  if (is.null(ctrl$nnzlmax)) 
    nnzlmax <- 4 * nnzdmax
  wwm <- vector("numeric", 3 * m)
  s <- u - x
  b1 <- solve(e, ao %*% y, tmpmax = tmpmax, nnzlmax = nnzlmax, 
              nsubmax = nsubmax)
  r <- y - a %*% b1
  z <- ifelse(abs(r) < ctrl$small, (r * (r > 0) + ctrl$small), 
              r * (r > 0))
  w <- z - r
  wwn <- matrix(0, n, 14)
  wwn[, 1] <- r
  wwn[, 2] <- z
  wwn[, 3] <- w
  srqfnb.o <- .Fortran("srqfn", n = as.integer(n), m = as.integer(m), 
                       nnza = as.integer(nnza), a = as.double(a@ra), ja = as.integer(a@ja), 
                       ia = as.integer(a@ia), ao = as.double(ao@ra), jao = as.integer(ao@ja), 
                       iao = as.integer(ao@ia), nnzdmax = as.integer(nnzdmax), 
                       d = double(nnzdmax), jd = integer(nnzdmax), id = integer(m + 
                                                                                  1), dsub = double(nnzemax + 1), jdsub = integer(nnzemax + 
                                                                                                                                    1), nnzemax = as.integer(nnzemax), e = as.double(e@ra), 
                       je = as.integer(e@ja), ie = as.integer(e@ia), nsubmax = as.integer(nsubmax), 
                       lindx = integer(nsubmax), xlindx = integer(m + 1), nnzlmax = as.integer(nnzlmax), 
                       lnz = double(nnzlmax), xlnz = integer(m + 1), iw = integer(m * 
                                                                                    5), iwmax = as.integer(iwmax), iwork = integer(iwmax), 
                       xsuper = integer(m + 1), tmpmax = as.integer(tmpmax), 
                       tmpvec = double(tmpmax), wwm = as.double(wwm), wwn = as.double(wwn), 
                       cachsz = as.integer(ctrl$cachsz), level = as.integer(8), 
                       x = as.double(x), s = as.double(s), u = as.double(u), 
                       c = as.double(y), sol = as.double(b1), rhs = as.double(rhs), 
                       small = as.double(ctrl$small), ierr = integer(1), maxiter = as.integer(ctrl$maxiter), 
                       time = double(7), PACKAGE = "quantreg")[c("sol", "ierr", 
                                                                 "maxiter", "time")]
  ierr <- srqfnb.o$ierr
  if (!(ierr == 0) && ctrl$warn.mesg) 
    warning(sfnMessage(ierr))          # change with respect to rq.fit.sfn
  coefficients <- -srqfnb.o$sol
  return(list(coefficients = coefficients, ierr = ierr, it = srqfnb.o$maxiter, 
       time = sum(srqfnb.o$time), tau = tau))
}
